<?php

declare(strict_types = 1);


namespace Drupal\rokka;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\image\Entity\ImageStyle;
use Drupal\rokka\Entity\RokkaStack;
use Drupal\rokka\StyleEffects\ImageEffectInterface;
use Exception;
use GuzzleHttp\Psr7\Uri;
use Rokka\Client\Core\Stack as RemoteStack;
use Rokka\Client\UriHelper;

/**
 * Image style and stack helper service.
 */
class StackHelper {

  use StringTranslationTrait;

  /**
   * Rokka service.
   *
   * @var \Drupal\rokka\RokkaServiceInterface
   */
  protected $rokkaService;

  /**
   * Rokka stack storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $storageRokkaStack;

  /**
   * Image style storage.
   *
   * @var \Drupal\image\ImageStyleStorage
   */
  protected $storageImageStyle;

  /**
   * Logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * StackHelper constructor.
   *
   * @param \Drupal\rokka\RokkaServiceInterface $rokkaService
   *   Rokka service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity type manager.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   Logger.
   */
  public function __construct(
    RokkaServiceInterface $rokkaService,
    EntityTypeManagerInterface $entityTypeManager,
    LoggerChannelInterface $logger
  ) {
    $this->rokkaService = $rokkaService;
    $this->storageRokkaStack = $entityTypeManager->getStorage('rokka_stack');
    $this->storageImageStyle = $entityTypeManager->getStorage('image_style');
    $this->logger = $logger;
  }

  /**
   * Build stack operation collection.
   *
   * @param array $effects
   *   Effects.
   *
   * @return \Rokka\Client\Core\StackOperation[]
   *   Stack operations.
   */
  private static function buildStackOperationCollection(array $effects): ?array {
    if (empty($effects)) {
      $effects = [
        [
          'id' => 'noop',
          'data' => [],
          'weight' => 0,
        ],
      ];
    }

    $operations = [];
    $currentId = 0;

    // Sort effects by defined weight.
    usort($effects,
      static function ($a, $b) {
        return $a['weight'] - $b['weight'];
      }
    );

    foreach ($effects as $effect) {
      $options = self::buildStackOperation($effect);
      if (!empty($options)) {
        foreach ($options as $option) {
          $operations[$currentId++] = $option;
        }
      }
    }

    if (empty($operations)) {
      return NULL;
    }

    ksort($operations);
    return $operations;
  }

  /**
   * Build stack operation.
   *
   * @param array $effect
   *   Effects.
   *
   * @return \Rokka\Client\Core\StackOperation[]
   *   Operations.
   */
  public static function buildStackOperation(array $effect): array {
    $name = $effect['id'];
    $class_name = 'Drupal\rokka\StyleEffects\Effect' . self::convertToCamelCase(
        $name,
        TRUE
      );

    $result = [];
    if (
      class_exists($class_name) &&
      in_array(ImageEffectInterface::class, class_implements($class_name), TRUE)
    ) {
      /** @var \Drupal\rokka\StyleEffects\ImageEffectInterface $class_name */
      $result = $class_name::buildRokkaStackOperation(!empty($effect['data']) ? $effect['data'] : []);
    }
    else {
      // TODO: Inject.
      \Drupal::logger('rokka')->error(
        'Can not convert effect "%effect" to Rokka.io StackOperation: "%class" Class missing!',
        [
          '%effect' => $name,
          '%class' => $class_name,
        ]
      );
    }

    return $result;
  }

  /**
   * Get effect stack.
   *
   * @param \Drupal\image\Entity\ImageStyle|null $image_style
   *   Image style.
   *
   * @return \Rokka\Client\Core\StackOperation[]
   *   Stack operations.
   */
  public static function getEffectStack(ImageStyle $image_style = NULL): ?array {
    $result = [];

    if ($image_style !== NULL) {
      $collection = self::buildStackOperationCollection($image_style->getEffects()->getConfiguration());
      if (!empty($collection)) {
        $result = $collection;
      }
    }

    return $result;
  }

  /**
   * Camel Case convert.
   *
   * @param string $subject
   *   Input string.
   * @param bool $class_case
   *   Class case.
   *
   * @return string
   *   Output string.
   */
  public static function convertToCamelCase(
    string $subject,
    $class_case = FALSE
  ): string {
    // non-alpha and non-numeric characters become spaces.
    $subject = preg_replace('/[^a-z0-9]+/i', ' ', $subject);
    $subject = trim($subject);
    // Uppercase the first character of each word.
    $subject = ucwords($subject);
    $subject = str_replace(' ', '', $subject);
    if (!$class_case) {
      $subject = lcfirst($subject);
    }

    return $subject;
  }

  /**
   * Image style sync.
   *
   * @param \Drupal\image\Entity\ImageStyle $entity
   *   Image style.
   */
  public function imageStyleSync(ImageStyle $entity): void {

    $client = $this->rokkaService->getRokkaImageClient();

    $effect_stack = self::getEffectStack($entity);

    [$stackOptions, $outputFormat] = $this->normalizeFormParameters($entity);

    $stack = new RemoteStack(
      $this->rokkaService->getRokkaOrganizationName(),
      $entity->getName(),
      $effect_stack,
      $stackOptions
    );
    try {
      $remoteStack = $client->saveStack($stack, ['overwrite' => TRUE]);
    }
    catch (\Exception $e) {
      $this->logger->error(
        'Cannot save remote Rokka stack @name',
        ['@name' => $entity->getName()]
      );
      return;
    }

    $rokkaStack = $this->storageRokkaStack->load($entity->getName());
    if ($rokkaStack instanceof RokkaStack) {
      $this->logger->debug(
        'Updating existing Rokka stack @name',
        ['@name' => $entity->getName()]
      );
      $rokkaStack->setStackOptions($remoteStack->stackOptions);
      $rokkaStack->setOrganization($remoteStack->organization);
      $rokkaStack->setOutputFormat($outputFormat);
    }
    else {
      $this->logger->debug(
        'Creating new Rokka stack @name',
        ['@name' => $entity->getName()]
      );
      $rokkaStack = $this->storageRokkaStack->create(
        [
          'id' => $remoteStack->name,
          'organization' => $remoteStack->organization,
          'stackOperations' => $remoteStack->stackOperations,
          'outputFormat' => $outputFormat,
          'label' => $remoteStack->name,
        ]
      );
      $rokkaStack->setStackOptions($remoteStack->stackOptions);
    }
    $rokkaStack->enforceIsNew(FALSE);
    $rokkaStack->save();

    // Flush the image style cache.
    $entity->flush();
  }

  /**
   * Image style preview.
   *
   * @param array $variables
   *   Variables.
   */
  public function imageStylePreview(array &$variables): void {
    $previewUrl = $variables['derivative']['url'];
    $previewUri = new Uri($previewUrl);
    $components = UriHelper::decomposeUri($previewUri);

    if ($components === NULL) {
      return;
    }

    $stackUri = $components->getStack();
    $imageStyle = $this->storageImageStyle->load($stackUri->getName());
    $stackEntity = $this->storageRokkaStack->load($stackUri->getName());

    $effectStack = self::getEffectStack($imageStyle);
    $stack = new RemoteStack(NULL, $components['stack'], $effectStack);
    if ($stackEntity instanceof RokkaStack) {
      $stack_options = $stackEntity->getStackOptions();
      if (is_array($stack_options)) {
        $stack->setStackOptions($stack_options);
      }
    }
    $dynamic_stack = $stack->getDynamicUriString();
    $components->setStack($dynamic_stack);
    $previewUrl = UriHelper::composeUri($components, $previewUri);
    $variables['derivative']['rendered']['#uri'] = (string) $previewUrl;
    $variables['derivative']['url'] = (string) $previewUrl;
    // We have to get the image from rokka to know its dimensions
    // that makes is a little bit slower, if anyone has a better idea..
    $image = @file_get_contents($previewUrl);
    if ($image !== FALSE) {
      $dimensions = $this->getImageDimensionsFromContent($image);
      $variables['derivative']['width'] = $dimensions['width'];
      $variables['derivative']['height'] = $dimensions['height'];
    }
    else {
      $variables['derivative']['width'] = 'unknown ';
      $variables['derivative']['height'] = 'unknown ';
    }
  }

  /**
   * Gets image dimensions via finfo.
   *
   * Finfo is a pretty fast way to get the dimensions of an image, unfortunately
   * it doesn't do that in an easy accessible way. We do it here from the
   * FILEINFO_RAW string and some regex magic.
   *
   * @param string $content
   *   Content.
   *
   * @return array
   *   Image dimensions.
   */
  private function getImageDimensionsFromContent(string $content): array {
    $answer = finfo_buffer(finfo_open(FILEINFO_RAW), $content);
    if (preg_match('/, *([0-9]{1,6}) *x *([0-9]{1,6})/', $answer, $matches)) {
      return ['width' => (int) $matches[1], 'height' => (int) $matches[2]];
    }

    return ['width' => NULL, 'height' => NULL];
  }

  /**
   * Delete stack on image style delete.
   *
   * @param \Drupal\image\Entity\ImageStyle $entity
   *   Image style.
   */
  public function deleteStackOnStyleDelete(ImageStyle $entity): void {
    /** @var \Drupal\rokka\RokkaService $rokka_service */
    $rokka_service = \Drupal::service('rokka.service');
    $client = $rokka_service->getRokkaImageClient();

    try {
      \Drupal::configFactory()
        ->getEditable('rokka.rokka_stack.' . $entity->getName())
        ->delete();
      $client->deleteStack($entity->getName());
    }
    catch (Exception $e) {
      // If the Stack is not found, that means that it is already deleted.
      if ($e->getCode() !== 404) {
        \Drupal::logger('rokka')->error(
          'Exception while deleting ImageStack "@name" in "@function". Error: "@error"',
          [
            '@name' => $entity->getName(),
            '@function' => __FUNCTION__,
            '@error' => $e->getMessage(),
          ]
        );
      }
    }
  }

  /**
   * Alter the image style form.
   *
   * @param array $form
   *   Form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state interface.
   */
  public function alterImageStyleForm(array &$form, FormStateInterface $form_state): void {
    /** @var \Drupal\image\Form\ImageStyleEditForm $form_object */
    $form_object = $form_state->getFormObject();
    // TODO: Inject.
    $config = \Drupal::config('rokka.settings');

    /** @var \Drupal\rokka\Entity\RokkaStack $stack_entity */
    $stack_entity = NULL;
    // Can be empty, if we create a new image style.
    if (!empty($form_object->getEntity()->get('name'))) {
      $stack_entity = $this->storageRokkaStack->load(
          $form_object->getEntity()->get('name')
      );
    }

    if ($stack_entity instanceof RokkaStack) {
      $stack_options = $stack_entity->getStackOptions();
      $default_format = $stack_entity->getOutputFormat();
    }
    else {
      $stack_options = [];
      $default_format = $config->get('output_format');
    }

    $user_input = $form_state->getUserInput();
    $form['rokka_stack_options'] = [
      '#type' => 'fieldset',
      '#title' => 'Rokka Stack settings',
      '#collapsible' => TRUE,
      '#collapsed' => FALSE,
      '#tree' => TRUE,
    ];

    $form['rokka_stack_options']['rokka_output_format'] = [
      '#type' => 'select',
      '#title' => $this->t('Output format'),
      '#description' => $this->t('Defines the delivered image format.'),
      '#required' => TRUE,
      '#default_value' => $default_format,
      '#options' => [
        'jpg' => $this->t('JPG'),
        'png' => $this->t('PNG'),
        'gif' => $this->t('GIF'),
      ],
    ];

    $form['rokka_stack_options']['rokka_jpg_quality'] = [
      '#type' => 'textfield',
      '#title' => $this->t('JPG quality'),
      '#description' => $this->t(
        'JPEG Quality: from 1 (high compression, low quality) to 100 (low compression, high quality)'
      ),
      '#size' => 20,
      '#maxlength' => 3,
      '#required' => FALSE,
      '#default_value' => $config->get('jpg_quality') ?? 0,
      '#min' => 0,
      '#max' => 100,
    ];
    self::setValueForField(
      'rokka_jpg_quality',
      'jpg.quality',
      $form,
      $user_input,
      $stack_options
    );

    $form['rokka_stack_options']['rokka_webp_quality'] = [
      '#type' => 'textfield',
      '#title' => $this->t('WEBP quality'),
      '#description' => $this->t(
        'WEBP Quality: from 1 (high compression, low quality) to 100 (low compression, high quality)'
      ),
      '#size' => 20,
      '#maxlength' => 3,
      '#required' => FALSE,
      '#default_value' => $config->get('webp_quality') ?? 0,
      '#min' => 0,
      '#max' => 100,
    ];

    self::setValueForField(
      'rokka_webp_quality',
      'webp.quality',
      $form,
      $user_input,
      $stack_options
    );

    $form['rokka_stack_options']['rokka_autoformat'] = [
      '#type' => 'radios',
      '#title' => $this->t('Autoformat'),
      '#description' => $this->t(
        'If set, rokka delivers the best format supported by the browser.'
      ),
      '#required' => FALSE,
      '#default_value' => $config->get('autoformat') ?? 'none',
      '#options' => [
        'true' => $this->t('True'),
        'false' => $this->t('False'),
        'none' => $this->t('Rokka default (false)'),
      ],

    ];

    self::setValueForField(
      'rokka_autoformat',
      'autoformat',
      $form,
      $user_input,
      $stack_options
    );
  }

  /**
   * Set value for field.
   *
   * @param string $form_field_name
   *   Source field name.
   * @param string $stack_option_name
   *   Stack option name.
   * @param array $form
   *   Form.
   * @param array $user_input
   *   User input.
   * @param array $stack_options
   *   Stack options.
   *
   * @FIXME: Remove this function, unhelpful obfuscation.
   */
  public static function setValueForField(
    string $form_field_name,
    string $stack_option_name,
    array &$form,
    array $user_input,
    array $stack_options
  ): void {
    if (isset($user_input['rokka_stack_options'][$form_field_name])) {
      $form['rokka_stack_options'][$form_field_name]['#value'] = $user_input['rokka_stack_options'][$form_field_name];
    }
    elseif (isset($stack_options[$stack_option_name])) {
      if ($stack_options[$stack_option_name] === TRUE) {
        $form['rokka_stack_options'][$form_field_name]['#value'] = 'true';
      }
      elseif ($stack_options[$stack_option_name] === FALSE) {
        $form['rokka_stack_options'][$form_field_name]['#value'] = 'false';
      }
      else {
        $form['rokka_stack_options'][$form_field_name]['#value'] = $stack_options[$stack_option_name];

      }
    }
    if (isset($form['rokka_stack_options'][$form_field_name]['#value'])) {
      $form['rokka_stack_options'][$form_field_name]['#default_value'] = $form['rokka_stack_options'][$form_field_name]['#value'];
    }
  }

  /**
   * Normalize saved / form values for consistent usage.
   *
   * Can be simplified / cleaned up.
   *
   * @param \Drupal\image\Entity\ImageStyle $entity
   *   Entity.
   *
   * @return array
   *   Stack options and output format..
   */
  protected function normalizeFormParameters(ImageStyle $entity): array {
    // TODO: Inject.
    $config = \Drupal::config('rokka.settings');

    // Create the stack on rokka.
    $stackOptionsForm = $entity->get('rokka_stack_options');
    if (empty($stackOptionsForm)) {
      // Fallback to stack data.
      $rokkaStack = $this->rokkaService->loadStackByName($entity->getName());
      if ($rokkaStack instanceof RokkaStack) {
        if (!empty($rokkaStack->getStackOptions()['jpg.quality'])) {
          $stackOptionsForm['rokka_jpg_quality'] = $rokkaStack->getStackOptions(
          )['jpg.quality'];
        }
        if (!empty($rokkaStack->getStackOptions()['webp.quality'])) {
          $stackOptionsForm['rokka_webp_quality'] = $rokkaStack->getStackOptions(
          )['webp.quality'];
        }
        $stackOptionsForm['rokka_output_format'] = $rokkaStack->getOutputFormat(
        );
      }
    }

    $stackOptions = [];

    if (!empty($stackOptionsForm['rokka_jpg_quality']) && (integer) $stackOptionsForm['rokka_jpg_quality'] >= 1) {
      $stackOptions['jpg.quality'] = (integer) $stackOptionsForm['rokka_jpg_quality'];
    }
    elseif ((int) $config->get('jpg_quality') >= 1) {
      $stackOptions['jpg.quality'] = $config->get('jpg_quality');
    }

    if (!empty($stackOptionsForm['rokka_webp_quality']) && (integer) $stackOptionsForm['rokka_webp_quality'] >= 1) {
      $stackOptions['webp.quality'] = (integer) $stackOptionsForm['rokka_webp_quality'];
    }
    elseif ((int) $config->get('webp_quality') >= 1) {
      $stackOptions['webp.quality'] = $config->get('webp_quality');
    }

    // TODO: Possible bug here resetting autoformat on creation from style.
    $stackOptions['autoformat'] = $config->get('autoformat');
    if (!empty($stackOptionsForm['rokka_autoformat'])) {
      $stackOptions['autoformat'] = $stackOptionsForm['rokka_autoformat'];
    }
    if (
      !empty($stackOptionsForm['rokka_autoformat']) &&
      ($stackOptionsForm['rokka_autoformat'] === 'none' || $stackOptionsForm['rokka_autoformat'] === NULL)
    ) {
      unset($stackOptions['autoformat']);
    }

    $outputFormat = $config->get('output_format');
    if (!empty($stackOptionsForm['rokka_output_format'])) {
      $outputFormat = $stackOptionsForm['rokka_output_format'];
    }
    return [$stackOptions, $outputFormat];
  }

}
